@@ -1,8 +1,12 @@
-{-# LANGUAGE TemplateHaskell #-}
 {-# LANGUAGE CPP #-}
+#if WITH_TEMPLATE_HASKELL
+{-# LANGUAGE TemplateHaskell #-}
+#endif
 {-# LANGUAGE TypeFamilies #-}
+{-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
+{-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE UndecidableInstances #-}
 -- |  This module provides the facilities needed for a decoupled logging system.
 --
@@ -26,7 +30,9 @@ module Control.Monad.Logger
     , LoggingT (..)
     , runStderrLoggingT
     , runStdoutLoggingT
+    , withChannelLogger
     , NoLoggingT (..)
+#if WITH_TEMPLATE_HASKELL
     -- * TH logging
     , logDebug
     , logInfo
@@ -41,26 +47,51 @@ module Control.Monad.Logger
     , logOtherS
     -- * TH util
     , liftLoc
+#endif
+    -- * Non-TH logging
+    , logDebugN
+    , logInfoN
+    , logWarnN
+    , logErrorN
+    , logOtherN
+    -- * Non-TH logging with source
+    , logDebugNS
+    , logInfoNS
+    , logWarnNS
+    , logErrorNS
+    , logOtherNS
+
+    -- * utilities for defining your own loggers
+    , defaultLogStr
+    , Loc
     ) where
 
+#if WITH_TEMPLATE_HASKELL
 import Language.Haskell.TH.Syntax (Lift (lift), Q, Exp, Loc (..), qLocation)
-import System.Log.FastLogger (ToLogStr (toLogStr), LogStr (..))
+#endif
 
 import Data.Monoid (Monoid)
 
 import Control.Applicative (Applicative (..))
-import Control.Monad (liftM, ap)
+import Control.Concurrent.STM
+import Control.Concurrent.STM.TBChan
+import Control.Exception.Lifted (onException)
+import Control.Monad (liftM, ap, when, void)
 import Control.Monad.Base (MonadBase (liftBase))
+import Control.Monad.Loops (untilM)
 import Control.Monad.Trans.Control (MonadBaseControl (..), MonadTransControl (..))
-import Data.Functor.Identity (Identity)
-import Control.Monad.ST (ST)
-import qualified Control.Monad.ST.Lazy as Lazy (ST)
 import qualified Control.Monad.Trans.Class as Trans
 
-import System.IO (stdout, stderr, Handle)
-
 import Control.Monad.IO.Class (MonadIO (liftIO))
-import Control.Monad.Trans.Resource (MonadResource (liftResourceT), MonadThrow (monadThrow))
+import Control.Monad.Trans.Resource (MonadResource (liftResourceT), MonadThrow, monadThrow)
+#if MIN_VERSION_resourcet(1,1,0)
+import Control.Monad.Trans.Resource (throwM)
+import Control.Monad.Catch (MonadCatch (..)
+#if MIN_VERSION_exceptions(0,6,0)
+    , MonadMask (..)
+#endif
+    )
+#endif
 
 import Control.Monad.Trans.Identity ( IdentityT)
 import Control.Monad.Trans.List     ( ListT    )
@@ -78,10 +109,13 @@ import qualified Control.Monad.Trans.RWS.Strict    as Strict ( RWST   )
 import qualified Control.Monad.Trans.State.Strict  as Strict ( StateT )
 import qualified Control.Monad.Trans.Writer.Strict as Strict ( WriterT )
 
-import Data.Text (Text, pack, unpack, empty)
+import Data.Text (Text, pack, unpack)
 import qualified Data.Text as T
 import qualified Data.ByteString.Char8 as S8
-import Data.Text.Encoding (encodeUtf8)
+
+import Data.Monoid (mappend, mempty)
+import System.Log.FastLogger
+import System.IO (Handle, stdout, stderr)
 
 import Control.Monad.Cont.Class   ( MonadCont (..) )
 import Control.Monad.Error.Class  ( MonadError (..) )
@@ -90,21 +124,49 @@ import Control.Monad.Reader.Class ( MonadReader (..) )
 import Control.Monad.State.Class  ( MonadState (..) )
 import Control.Monad.Writer.Class ( MonadWriter (..) )
 
+import Blaze.ByteString.Builder (toByteString)
+
+import Prelude hiding (catch)
+
+#if !MIN_VERSION_fast_logger(2, 1, 0) && MIN_VERSION_bytestring(0, 10, 2)
+import qualified Data.ByteString.Lazy as L
+import Data.ByteString.Builder (toLazyByteString)
+#endif
+
+#if MIN_VERSION_conduit_extra(1,1,0)
+import Data.Conduit.Lazy (MonadActive, monadActive)
+#endif
+
 data LogLevel = LevelDebug | LevelInfo | LevelWarn | LevelError | LevelOther Text
     deriving (Eq, Prelude.Show, Prelude.Read, Ord)
 
+type LogSource = Text
+
+#if WITH_TEMPLATE_HASKELL
+
 instance Lift LogLevel where
     lift LevelDebug = [|LevelDebug|]
-    lift LevelInfo = [|LevelInfo|]
-    lift LevelWarn = [|LevelWarn|]
+    lift LevelInfo  = [|LevelInfo|]
+    lift LevelWarn  = [|LevelWarn|]
     lift LevelError = [|LevelError|]
     lift (LevelOther x) = [|LevelOther $ pack $(lift $ unpack x)|]
 
-type LogSource = Text
+#else
+
+data Loc
+  = Loc { loc_filename :: String
+    , loc_package  :: String
+    , loc_module   :: String
+    , loc_start    :: CharPos
+    , loc_end      :: CharPos }
+type CharPos = (Int, Int)
+
+#endif
 
 class Monad m => MonadLogger m where
     monadLoggerLog :: ToLogStr msg => Loc -> LogSource -> LogLevel -> msg -> m ()
 
+
 {-
 instance MonadLogger IO          where monadLoggerLog _ _ _ = return ()
 instance MonadLogger Identity    where monadLoggerLog _ _ _ = return ()
@@ -130,6 +192,7 @@ instance (MonadLogger m, Monoid w) => MonadLogger (Strict.WriterT w m) where DEF
 instance (MonadLogger m, Monoid w) => MonadLogger (Strict.RWST r w s m) where DEF
 #undef DEF
 
+#if WITH_TEMPLATE_HASKELL
 logTH :: LogLevel -> Q Exp
 logTH level =
     [|monadLoggerLog $(qLocation >>= liftLoc) (pack "") $(lift level) . (id :: Text -> Text)|]
@@ -170,7 +233,7 @@ liftLoc (Loc a b c (d1, d2) (e1, e2)) = [|Loc
 
 -- | Generates a function that takes a 'LogSource' and 'Text' and logs a 'LevelDebug' message. Usage:
 --
--- > $logDebug "SomeSource" "This is a debug log message"
+-- > $logDebugS "SomeSource" "This is a debug log message"
 logDebugS :: Q Exp
 logDebugS = [|\a b -> monadLoggerLog $(qLocation >>= liftLoc) a LevelDebug (b :: Text)|]
 
@@ -186,9 +249,10 @@ logErrorS = [|\a b -> monadLoggerLog $(qLocation >>= liftLoc) a LevelError (b ::
 
 -- | Generates a function that takes a 'LogSource', a level name and a 'Text' and logs a 'LevelOther' message. Usage:
 --
--- > $logOther "SomeSource" "My new level" "This is a log message"
+-- > $logOtherS "SomeSource" "My new level" "This is a log message"
 logOtherS :: Q Exp
 logOtherS = [|\src level msg -> monadLoggerLog $(qLocation >>= liftLoc) src (LevelOther level) (msg :: Text)|]
+#endif
 
 -- | Monad transformer that disables logging.
 --
@@ -209,8 +273,32 @@ instance Monad m => Monad (NoLoggingT m) where
 instance MonadIO m => MonadIO (NoLoggingT m) where
     liftIO = Trans.lift . liftIO
 
+#if MIN_VERSION_resourcet(1,1,0)
+instance MonadThrow m => MonadThrow (NoLoggingT m) where
+    throwM = Trans.lift . throwM
+
+instance MonadCatch m => MonadCatch (NoLoggingT m) where
+    catch (NoLoggingT m) c =
+        NoLoggingT $ m `catch` \e -> runNoLoggingT (c e)
+#if MIN_VERSION_exceptions(0,6,0)
+instance MonadMask m => MonadMask (NoLoggingT m) where
+#endif
+    mask a = NoLoggingT $ mask $ \u -> runNoLoggingT (a $ q u)
+      where q u (NoLoggingT b) = NoLoggingT $ u b
+    uninterruptibleMask a =
+        NoLoggingT $ uninterruptibleMask $ \u -> runNoLoggingT (a $ q u)
+      where q u (NoLoggingT b) = NoLoggingT $ u b
+#else
 instance MonadThrow m => MonadThrow (NoLoggingT m) where
     monadThrow = Trans.lift . monadThrow
+#endif
+
+#if MIN_VERSION_conduit_extra(1,1,0)
+instance MonadActive m => MonadActive (NoLoggingT m) where
+    monadActive = Trans.lift monadActive
+instance MonadActive m => MonadActive (LoggingT m) where
+    monadActive = Trans.lift monadActive
+#endif
 
 instance MonadResource m => MonadResource (NoLoggingT m) where
     liftResourceT = Trans.lift . liftResourceT
@@ -260,8 +348,24 @@ instance Monad m => Monad (LoggingT m) where
 instance MonadIO m => MonadIO (LoggingT m) where
     liftIO = Trans.lift . liftIO
 
+#if MIN_VERSION_resourcet(1,1,0)
+instance MonadThrow m => MonadThrow (LoggingT m) where
+    throwM = Trans.lift . throwM
+instance MonadCatch m => MonadCatch (LoggingT m) where
+  catch (LoggingT m) c =
+      LoggingT $ \r -> m r `catch` \e -> runLoggingT (c e) r
+#if MIN_VERSION_exceptions(0,6,0)
+instance MonadMask m => MonadMask (LoggingT m) where
+#endif
+  mask a = LoggingT $ \e -> mask $ \u -> runLoggingT (a $ q u) e
+    where q u (LoggingT b) = LoggingT (u . b)
+  uninterruptibleMask a =
+    LoggingT $ \e -> uninterruptibleMask $ \u -> runLoggingT (a $ q u) e
+      where q u (LoggingT b) = LoggingT (u . b)
+#else
 instance MonadThrow m => MonadThrow (LoggingT m) where
     monadThrow = Trans.lift . monadThrow
+#endif
 
 instance MonadResource m => MonadResource (LoggingT m) where
     liftResourceT = Trans.lift . liftResourceT
@@ -296,9 +400,52 @@ defaultOutput :: Handle
               -> LogStr
               -> IO ()
 defaultOutput h loc src level msg =
-    S8.hPutStrLn h $ S8.concat bs
+    S8.hPutStrLn h ls
   where
-    bs =
+    ls = defaultLogStrBS loc src level msg
+defaultLogStrBS :: Loc
+                -> LogSource
+                -> LogLevel
+                -> LogStr
+                -> S8.ByteString
+defaultLogStrBS a b c d =
+    toBS $ defaultLogStr a b c d
+  where
+#if MIN_VERSION_fast_logger(2, 1, 0)
+    toBS = fromLogStr
+#elif MIN_VERSION_bytestring(0, 10, 2)
+    toBS = L.toStrict . toLazyByteString . logStrBuilder
+#else
+    toBS = toByteString . logStrBuilder
+#endif
+
+defaultLogLevelStr :: LogLevel -> LogStr
+defaultLogLevelStr level = case level of
+    LevelOther t -> toLogStr t
+    _            -> toLogStr $ S8.pack $ drop 5 $ show level
+
+defaultLogStr :: Loc
+              -> LogSource
+              -> LogLevel
+              -> LogStr
+#if MIN_VERSION_fast_logger(0, 2, 0)
+              -> LogStr
+#else
+              -> S8.ByteString
+#endif
+defaultLogStr loc src level msg =
+#if MIN_VERSION_fast_logger(0, 2, 0)
+    "[" `mappend` defaultLogLevelStr level `mappend`
+    (if T.null src
+        then mempty
+        else "#" `mappend` toLogStr src) `mappend`
+    "] " `mappend`
+    msg `mappend`
+    " @(" `mappend`
+    toLogStr (S8.pack fileLocStr) `mappend`
+    ")\n"
+#else
+    S8.concat
         [ S8.pack "["
         , case level of
             LevelOther t -> encodeUtf8 t
@@ -314,7 +461,8 @@ defaultOutput h loc src level msg =
         , encodeUtf8 $ pack fileLocStr
         , S8.pack ")\n"
         ]
-
+#endif
+  where
     -- taken from file-location package
     -- turn the TH Loc loaction information into a human readable string
     -- leaving out the loc_end parameter
@@ -323,6 +471,18 @@ defaultOutput h loc src level msg =
       where
         line = show . fst . loc_start
         char = show . snd . loc_start
+{-
+defaultLogStrWithoutLoc ::
+    LogSource -> LogLevel -> LogStr -> LogStr
+defaultLogStrWithoutLoc loc src level msg =
+    "[" `mappend` defaultLogLevelStr level `mappend`
+    (if T.null src
+        then mempty
+        else "#" `mappend` toLogStr src) `mappend`
+    "] " `mappend`
+    msg `mappend` "\n"
+-}
+
 
 -- | Run a block using a @MonadLogger@ instance which prints to stderr.
 --
@@ -336,6 +496,27 @@ runStderrLoggingT = (`runLoggingT` defaultOutput stderr)
 runStdoutLoggingT :: MonadIO m => LoggingT m a -> m a
 runStdoutLoggingT = (`runLoggingT` defaultOutput stdout)
 
+-- | Within the 'LoggingT' monad, capture all log messages to a bounded
+--   channel of the indicated size, and only actually log them if there is an
+--   exception.
+--
+-- Since 0.3.2
+withChannelLogger :: (MonadBaseControl IO m, MonadIO m)
+                  => Int         -- ^ Number of mesasges to keep
+                  -> LoggingT m a
+                  -> LoggingT m a
+withChannelLogger size action = LoggingT $ \logger -> do
+    chan <- liftIO $ newTBChanIO size
+    runLoggingT action (channelLogger chan logger) `onException` dumpLogs chan
+  where
+    channelLogger chan logger loc src lvl str = atomically $ do
+        full <- isFullTBChan chan
+        when full $ void $ readTBChan chan
+        writeTBChan chan $ logger loc src lvl str
+
+    dumpLogs chan = liftIO $
+        sequence_ =<< atomically (untilM (readTBChan chan) (isEmptyTBChan chan))
+
 instance MonadCont m => MonadCont (LoggingT m) where
   callCC f = LoggingT $ \i -> callCC $ \c -> runLoggingT (f (LoggingT . const . c)) i
 
@@ -360,3 +541,39 @@ instance MonadWriter w m => MonadWriter w (LoggingT m) where
   tell   = Trans.lift . tell
   listen = mapLoggingT listen
   pass   = mapLoggingT pass
+
+defaultLoc :: Loc
+defaultLoc = Loc "<unknown>" "<unknown>" "<unknown>" (0,0) (0,0)
+
+logWithoutLoc :: (MonadLogger m, ToLogStr msg) => LogSource -> LogLevel -> msg -> m ()
+logWithoutLoc = monadLoggerLog defaultLoc
+
+logDebugN :: MonadLogger m => Text -> m ()
+logDebugN = logWithoutLoc "" LevelDebug
+
+logInfoN :: MonadLogger m => Text -> m ()
+logInfoN = logWithoutLoc "" LevelInfo
+
+logWarnN :: MonadLogger m => Text -> m ()
+logWarnN = logWithoutLoc "" LevelWarn
+
+logErrorN :: MonadLogger m => Text -> m ()
+logErrorN = logWithoutLoc "" LevelError
+
+logOtherN :: MonadLogger m => LogLevel -> Text -> m ()
+logOtherN = logWithoutLoc ""
+
+logDebugNS :: MonadLogger m => Text -> Text -> m ()
+logDebugNS src = logWithoutLoc src LevelDebug
+
+logInfoNS :: MonadLogger m => Text -> Text -> m ()
+logInfoNS src = logWithoutLoc src LevelInfo
+
+logWarnNS :: MonadLogger m => Text -> Text -> m ()
+logWarnNS src = logWithoutLoc src LevelWarn
+
+logErrorNS :: MonadLogger m => Text -> Text -> m ()
+logErrorNS src = logWithoutLoc src LevelError
+
+logOtherNS :: MonadLogger m => Text -> LogLevel -> Text -> m ()
+logOtherNS = logWithoutLoc
